#include "Driver.h"
#include <math.h>
#include <limits.h>
#include <gsl/gsl_sf_psi.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_erf.h>
#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>

namespace po = boost::program_options;

Driver::Driver(void)
{
}


Driver::~Driver(void)
{
}

bool Driver::Exit()
{
	return true;
}
bool Driver::LoadLabel(string filelist, vector <string> &labelset)
{
	ifstream  fp(filelist.c_str());
	if (!fp)
	{
		return false;
	}
	string buf;
	while (getline(fp, buf))
	{
		labelset.push_back(buf);
	}
	return true;
}
bool Driver::SaveScore(const string sScore_Result, const vector <string> &labelset, const mat &scoreVec, bool bsort = false)
{
	ofstream  fp(sScore_Result.c_str());
	vector <string> label = labelset;
	if (!fp)
	{
		printf("Can't Open %s For Saving Score\n", sScore_Result.c_str());
		return false;
	}
	if (labelset.size() == 0)
	{
		for (int i = 1; i <= scoreVec.n_rows; i++)
		{
			int idx = i;
			label.push_back(boost::lexical_cast<std::string>(idx));
		}
	}
	if (bsort)
	{
		uvec label_indices = sort_index(scoreVec.col(0), "descend");
		int topone_index = label_indices(0);
		/*if (scoreVec(topone_index, 0) <= dcutoff && scoreVec(topone_index, 0) > 0)
		{
		fp << label[topone_index] << "\t";
		for (int j = 0; j < scoreVec.n_cols - 1; j++)
		{

		fp << scoreVec(topone_index, j) << "\t";
		}
		if (scoreVec.n_cols > 0)
		{
		fp << scoreVec(topone_index, scoreVec.n_cols - 1) << "\n";
		}
		fp.close();
		return true;
		}*/
		for (int i = 0; i < label_indices.n_elem; i++)
		{
			int idx = label_indices(i);
			if (scoreVec(idx, 0) > dcutoff)
			{
				fp << label[idx] << "\t";
				for (int j = 0; j < scoreVec.n_cols - 1; j++)
				{

					fp << scoreVec(idx, j) << "\t";
				}
				if (scoreVec.n_cols > 0)
				{
					fp << scoreVec(idx, scoreVec.n_cols - 1) << "\n";
				}
			}
			else
			{
				break;
			}
		}
	}
	else
	{
		for (int i = 0; i < scoreVec.n_rows; i++)
		{
			int idx = i;
			if (scoreVec(idx, 0) > dcutoff)
			{
				fp << label[idx] << "\t";
				for (int j = 0; j < scoreVec.n_cols - 1; j++)
				{

					fp << scoreVec(idx, j) << "\t";
				}
				if (scoreVec.n_cols > 0)
				{
					fp << scoreVec(idx, scoreVec.n_cols - 1) << "\n";
				}
			}
			else
			{
				break;
			}
		}
		fp.close();
		return true;
	}

	fp.close();
	return true;
}

bool Driver::Compute_Cluster()
{
	ifstream  fp(sClusterList.c_str());
	if (!fp)
	{
		return false;
	}
	string buf;
	vector<string> outputs;
	vector<string> inputs;
	while (getline(fp, buf))
	{
		filesystem::path pathtmp(buf);
		string str = filesystem::basename(pathtmp);
		outputs.push_back(sCluster_Result + "/" + str + ".cluster");
		inputs.push_back(buf);
	}
	Load();
	ComputeConst();
#pragma omp parallel for
	for (int i = 0; i < inputs.size(); i++)
	{
		mat ProjectMat, labelProbMat;
		ProjectMat.load(inputs[i]);
		ProjectMat = ProjectMat.t();
		Cluster_Vector(ProjectMat, labelProbMat);
		SaveScore(outputs[i], labelset, labelProbMat, -10000);
	}
	return true;
}

bool Driver::Init(int ac, char* av[])

	//int nF, int nC_max, int nC_min, int uEM_ITERNUM, double uEM_MIN, double uWeightmin, double ustatsmin, int nCluster_Iternum_max, double uCluster_MIN, double pbeta0, double pbeta1, double pNu0, double palpha_s1, double palpha_s2)
{
	string config_file;
	string sModel;
	string sRunModel;
	//load config file and Command options
	try {
		//a group of options allowed only on command line
		po::options_description labelric("Command options");
		labelric.add_options()
			("version,v", "version information")
			("help", "help messages")
			("config,c", po::value<string>(&config_file)->default_value("iDriver.cfg"),
				"config file name")
			;
		po::options_description config("Configuration");
		config.add_options()
			("ThreadNum", po::value<int>(&threadnum)->default_value(1), "Thread Number")
			("Model", po::value<string>(&sModel)->default_value("DPM"), "Model Name")
			("Run_Mode", po::value<string>(&sRunModel)->default_value("train"), "Run Mode")
			("VecSize", po::value<int>(&D)->default_value(0), "Feature Vector Dim")
			("TruncateNum", po::value<int>(&K)->default_value(0), "Truncate Num")
			("TruncateMin", po::value<int>(&cluster_min)->default_value(1), "Truncate Num")
			("FeatureList", po::value<string>(&sFeatureList)->default_value(""), "labelMatrix File")
			("LabelFile", po::value<string>(&sLabelFile)->default_value(""), "labelIndex File")
			("Projection_Dir", po::value<string>(&sProjectionDir)->default_value(""), "Projection File")
			("Weight_min", po::value<double>(&dWeightmin)->default_value(1e-3), "Weight_min")
			("Weight_score", po::value<string>(&weight_str)->default_value("1"), "Weight_score")
			("Hp_Path", po::value<string>(&sHp_Path)->default_value("./output"), "Hyper-parameter Save Path")
			("beta0", po::value<double>(&vbeta0)->default_value(1.0), "beta0")
			("beta1", po::value<double>(&vbeta1)->default_value(12.0), "beta1")
			("alpha_s1", po::value<double>(&alpha_s1)->default_value(1.0), "alpha_s1")
			("alpha_s2", po::value<double>(&alpha_s2)->default_value(1.0), "alpha_s2")
			("Nu0", po::value<double>(&vNu0)->default_value(10.0), "Nu0 must >= D-1")
			("Score_Result", po::value<string>(&sScore_Dir)->default_value("result.score"), "Score Result")
			("cutoff", po::value<double>(&dcutoff)->default_value(-10000), "Score cutoff")
			("Cluster_Result", po::value<string>(&sCluster_Result)->default_value("result.cluster"), "Cluster Result")
			("Em_eps", po::value<double>(&eps)->default_value(0.01), "Em eps")
			("Em_Iternum_Max", po::value<int>(&iEM_Iternum_Max)->default_value(100), "Em Iter Max")
			;
		po::options_description cmdline_options;
		cmdline_options.add(labelric).add(config);
		po::options_description config_file_options;
		config_file_options.add(config);
		po::options_description visible("Allowed options");
		visible.add(labelric).add(config);
		po::variables_map vm;
		po::store(po::command_line_parser(ac, av).
			options(cmdline_options).run(), vm);
		po::notify(vm);

		ifstream ifs(config_file.c_str());
		if (!ifs)
		{
			cout << "can not open config file: " << config_file << "\n";
			return false;
		}
		else
		{
			po::store(po::parse_config_file(ifs, config_file_options), vm);
			po::notify(vm);
		}
		if (vm.count("help"))
		{
			cout << visible << "\n";
			return false;
		}
	}
	catch (std::exception& e)
	{
		cout << e.what() << "\n";
		return false;
	}
	transform(sModel.begin(), sModel.end(), sModel.begin(), ::tolower);
	if (sModel != "dpm")
	{
		return false;
	}
	transform(sRunModel.begin(), sRunModel.end(), sRunModel.begin(), ::tolower);
	if (sRunModel == "train")
	{
		mode = R_TRAIN;
	}
	else if (sRunModel == "projection")
	{
		mode = R_PROJECTION;
	}
	else if (sRunModel == "score")
	{
		mode = R_SCORE;
	}
	else if (sRunModel == "cluster")
	{
		mode = R_CLUSTER;
	}
	else if (sRunModel == "all")
	{
		mode = R_ALL;
	}
	LoadLabel(sLabelFile, labelset);
	Cluster_MIN = MIN_POST_PROB;
	EM_MIN = eps;
	ifstream  fp(sFeatureList.c_str());
	if (!fp)
	{
		return false;
	}
	string buf;
	vector <string> str_list;
	while (getline(fp, buf))
	{
		str_list.push_back(buf);
	}
	mat labelTrainMat;
	if (str_list.size() > 0)
	{
		labelTrainMat.load(str_list[0]);
	}
	D = labelTrainMat.n_cols;
	mM = zeros<mat>(D, K);
	mInvVar = zeros<cube>(D, D, K);
	vgamma1 = zeros<vec>(K);
	vgamma2 = zeros<vec>(K);
	vNu = vNu0 * ones<vec>(K);
	vbeta = vbeta0 * ones<vec>(K);
	dstatsmin = 1e-12;
	sClusterList = sProjectionList = sScoreList = sFeatureList;
	return true;
}

bool Driver::Load()
{
	FILE* fpModel;
	fpModel = fopen(sHp_Path.c_str(), "rb");
	if (fpModel == NULL)
	{
		printf("Error open %s for read.\n", sHp_Path.c_str());
		return false;
	}
	Model_Task mm;
	fread(&mm, 1, sizeof(Model_Task), fpModel);
	if (mm != T_DPM)
	{
		fclose(fpModel);
		return false;
	}
	fread(&K, 1, sizeof(int), fpModel);
	fread(&D, 1, sizeof(int), fpModel);
	fseek(fpModel, FILE_HEADER_SIZE, SEEK_SET);
	vWeight = zeros<vec>(K);
	mM = zeros<mat>(D, K);
	mInvVar = zeros<cube>(D, D, K);
	lndetW = zeros<vec>(K);
	vgamma1 = zeros<vec>(K);
	vgamma2 = zeros<vec>(K);
	vNu = zeros<vec>(K);
	vbeta = zeros<vec>(K);
	int nNumRead;
	nNumRead = fread(vWeight.memptr(), sizeof(double), vWeight.n_elem, fpModel);
	if (nNumRead != vWeight.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	nNumRead = fread(mM.memptr(), sizeof(double), mM.n_elem, fpModel);
	if (nNumRead != mM.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	nNumRead = fread(mInvVar.memptr(), sizeof(double), mInvVar.n_elem, fpModel);
	if (nNumRead != mInvVar.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	nNumRead = fread(lndetW.memptr(), sizeof(double), lndetW.n_elem, fpModel);
	if (nNumRead != lndetW.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	nNumRead = fread(vgamma1.memptr(), sizeof(double), vgamma1.n_elem, fpModel);
	if (nNumRead != vgamma1.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	nNumRead = fread(vgamma2.memptr(), sizeof(double), vgamma2.n_elem, fpModel);
	if (nNumRead != vgamma2.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	nNumRead = fread(vbeta.memptr(), sizeof(double), vbeta.n_elem, fpModel);
	if (nNumRead != vbeta.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	nNumRead = fread(vNu.memptr(), sizeof(double), vNu.n_elem, fpModel);
	if (nNumRead != vNu.n_elem)
	{
		fclose(fpModel);
		return false;
	}
	fread(&alpha_omega1, 1, sizeof(double), fpModel);
	fread(&alpha_omega2, 1, sizeof(double), fpModel);
	vMean = reshape(mM, mM.n_elem, 1);
	fclose(fpModel);
	return true;
}

bool Driver::Save()
{
	if (D <= 0)
	{
		printf("Error call WriteModel() : m_nVexSize<=0!");
		return false;
	}
	if (K <= 0)
	{
		printf("Error call WriteModel() : m_nMixNum<=0!");
		return false;
	}

	FILE* fpModel;
	Model_Task mm = T_DPM;
	fpModel = fopen(sHp_Path.c_str(), "wb");
	if (fpModel == NULL)
	{
		printf("Error open %s for read.\n", sHp_Path.c_str());
		return false;
	}
	fwrite(&mm, 1, sizeof(Model_Task), fpModel);
	fwrite(&K, 1, sizeof(int), fpModel);
	fwrite(&D, 1, sizeof(int), fpModel);
	fseek(fpModel, FILE_HEADER_SIZE, SEEK_SET);
	fwrite(vWeight.memptr(), sizeof(double), vWeight.n_elem, fpModel);
	fwrite(mM.memptr(), sizeof(double), mM.n_elem, fpModel);
	fwrite(mInvVar.memptr(), sizeof(double), mInvVar.n_elem, fpModel);
	fwrite(lndetW.memptr(), sizeof(double), lndetW.n_elem, fpModel);
	fwrite(vgamma1.memptr(), sizeof(double), vgamma1.n_elem, fpModel);
	fwrite(vgamma2.memptr(), sizeof(double), vgamma2.n_elem, fpModel);
	fwrite(vbeta.memptr(), sizeof(double), vbeta.n_elem, fpModel);
	fwrite(vNu.memptr(), sizeof(double), vNu.n_elem, fpModel);
	fwrite(&alpha_omega1, 1, sizeof(double), fpModel);
	fwrite(&alpha_omega2, 1, sizeof(double), fpModel);
	fclose(fpModel);
	return true;
}

bool Driver::ComputeConst()
{
	vE_V = zeros<vec>(K);
	vE_ln_pi = zeros<vec>(K);
	vE_1_V = zeros<vec>(K);
	vE_ln_L = zeros<vec>(K);
	E_alpha = alpha_omega1 / alpha_omega2;
	E_lnalpha = gsl_sf_psi(alpha_omega1) - log(alpha_omega2);
	rnk0 = -0.5 * D * log(2 * M_PI) * ones<vec>(K);
	for (int k = 0; k < K; k++)
	{
		vE_V(k) = gsl_sf_psi(vgamma1(k)) - gsl_sf_psi(vgamma1(k) + vgamma2(k));
		vE_1_V(k) = gsl_sf_psi(vgamma2(k)) - gsl_sf_psi(vgamma1(k) + vgamma2(k));
	}
	for (int k = 0; k < K; k++)
	{
		vE_ln_pi(k) = vE_V(k);
		for (int i = 0; i < k; i++)
		{
			vE_ln_pi(k) += vE_1_V(i);
		}
		vE_ln_L(k) = E_ln_L(lndetW(k), vNu(k), D);
	}
	for (int k = 0; k < K; k++)
	{
		rnk0(k) -= 0.5 * D / vbeta(k);
		rnk0(k) += vE_ln_pi(k) + 0.5 * vE_ln_L(k);
	}
	return true;
}

double Driver::Compute_rnk(vec& rnk, const vec& feat) const
{
	// probability vector of each Gaussion mixture
	rnk = rnk0;
	for (int k = 0; k < K; k++)
	{
		rnk(k) -= 0.5 * (vNu(k) * Mahalanobis_Distance(feat, mM.col(k), mInvVar.slice(k)));
	}
	double prob = rnk(0);
	for (int i = 1; i < K; i++)
	{
		prob = log_add(prob, rnk(i));
	}
	for (int i = 0; i < K; i++)
	{
		rnk(i) = exp(rnk(i) - prob);
	}
	return prob; //total probability of each Gaussion mixture
}

bool Driver::Project_Vector(mat& oriMat, mat& vvector, mat& probs)
{
	vvector = zeros<mat>(K * D, oriMat.n_cols);
	probs = zeros<mat>(K, oriMat.n_cols);
	// MPI with openMP for Multi threading  
	for (int i = 0; i < oriMat.n_cols; i++)
	{
		vec rnk = zeros<vec>(K);
		mat F = zeros<mat>(D, K);
		vec feat = oriMat.col(i);
		mat projectMean = zeros<mat>(D, K);
		if (vbeta1 == 0.0) //fast calculate projectMean
		{

			for (int k = 0; k < K; k++)
			{
				projectMean.col(k) = feat;
			}
		}
		else
		{
			Compute_rnk(rnk, feat);
			for (int k = 0; k < K; k++)
			{
				F.col(k) = rnk(k) * feat;
				double dbeta = vbeta1 + rnk(k);
				if (dbeta < dWeightmin)
				{
					projectMean.col(k) = feat;
				}
				else
				{
					projectMean.col(k) = (vbeta1 * mM.col(k) + F.col(k)) / dbeta;
				}
			}
		}
		projectMean.reshape(projectMean.n_elem, 1);
		vvector.col(i) = projectMean;
		probs.col(i) = rnk;
	}
	return true;
}

bool Driver::Cluster_Vector(mat& oriMat, mat& probs)
{
	probs = zeros<mat>(oriMat.n_cols, K);
	// MPI with openMP for Multi threading  
	for (int i = 0; i < oriMat.n_cols; i++)
	{
		vec rnk = zeros<vec>(K);
		vec feat = oriMat.col(i);
		Compute_rnk(rnk, feat);
		probs.row(i) = rnk.t();
	}
	return true;
}

bool Driver::Compute_Project()
{
	ifstream  fp(sProjectionList.c_str());
	if (!fp)
	{
		return false;
	}
	string buf;
	vector<string> outputs;
	vector<string> inputs;
	while (getline(fp, buf))
	{
		filesystem::path pathtmp(buf);
		string str = filesystem::basename(pathtmp); //save path
		outputs.push_back(sProjectionDir + "/" + str + ".vec");
		inputs.push_back(buf);
	}
	Load();
	ComputeConst();
#pragma omp parallel for
	for (int i = 0; i < inputs.size(); i++)
	{
		mat ProjectMat, labelScoreMat, labelProbMat;
		ProjectMat.load(inputs[i]);
		ProjectMat = ProjectMat.t();
		Project_Vector(ProjectMat, labelScoreMat, labelProbMat);
		labelScoreMat.save(outputs[i].c_str());
	}
	return true;
}

bool Driver::PreTrain()
{
	return true;
}

bool Driver::InitVariables(mat& vectors)
{
	alpha_omega1 = alpha_s1;
	alpha_omega2 = alpha_s2;
	vgamma1 = ones<vec>(K);
	vgamma2 = ones<vec>(K) * alpha_omega1 / alpha_omega2;
	mM0 = mean(vectors, 1);
	mW0 = cov(vectors.t());
	mM0 = zeros<vec>(D);
	lndetW = zeros<vec>(K);
	lndetW0 = log_det_inv(mW0);
	minvW0 = inv(mW0);
	double det_tmp = log_det_inv(minvW0);
	for (int i = 0; i < K; i++)
	{
		mInvVar.slice(i) = minvW0;
		mM.col(i) = mM0;
		lndetW(i) = det_tmp;
	}

	return true;
}

double Driver::ComputeCost(vec& N, mat& F, cube& S, double q_z_at_v) const
{
	double p_x_at_z_u_L = 0.0;
	double p_u_L = 0.0;
	double p_z_at_v = 0.0;
	double p_v_at_alpha = 0.0;
	double p_alpha = 0.0;
	double q_u_L = 0.0;
	double q_v_at_alpha = 0.0;
	double q_alpha = 0.0;
	int D = F.n_rows;

	for (int k = 0; k < K; k++)
	{
		if (vWeight(k) < dWeightmin)
		{
			continue;
		}
		double nk = N(k);
		vec Exk = F.col(k);
		mat Sk = S.slice(k);
		p_x_at_z_u_L += nk * (vE_ln_L(k) - D / vbeta(k) - vNu(k) * trace(solve(mInvVar.slice(k), Sk)) - vNu(k) * Mahalanobis_Distance(Exk, mM.col(k), mInvVar.slice(k)) - D * log(2 * M_PI));
		p_u_L += D * log(vbeta0 / (2 * M_PI)) + vE_ln_L(k) - D * vbeta0 / vbeta(k) - vbeta0 * vNu(k) * Mahalanobis_Distance(mM.col(k), mM0, mInvVar.slice(k)) + (vNu0 - D - 1) * vE_ln_L(k) - vNu(k) * trace(solve(mInvVar.slice(k), minvW0));
		q_u_L += 0.5 * vE_ln_L(k) + 0.5 * D * log(vbeta(k) / (2 * M_PI)) - 0.5 * D - H(lndetW(k), vNu(k), D);
		p_z_at_v += nk * vE_ln_pi(k);
		p_v_at_alpha += (E_alpha - 1) * vE_1_V(k);
		q_v_at_alpha += (vgamma1(k) - 1) * vE_V(k) + (vgamma2(k) - 1) * vE_1_V(k) + gsl_sf_lngamma(vgamma1(k) + vgamma2(k)) - gsl_sf_lngamma(vgamma1(k)) - gsl_sf_lngamma(vgamma2(k));
	}
	p_x_at_z_u_L *= 0.5;
	p_u_L *= 0.5;
	p_u_L -= K * Minus_ln_B(lndetW0, vNu0, D);
	p_v_at_alpha += K * E_lnalpha;
	p_alpha = (alpha_s1 - 1) * E_lnalpha - alpha_s2 * E_alpha - gsl_sf_lngamma(alpha_s1) + alpha_s1 * log(alpha_s2);
	q_alpha = (alpha_omega1 - 1) * E_lnalpha - alpha_omega2 * E_alpha - gsl_sf_lngamma(alpha_omega1) + alpha_omega1 * log(alpha_omega2);
	return p_x_at_z_u_L + p_u_L - q_u_L + p_z_at_v - q_z_at_v + p_v_at_alpha - q_v_at_alpha + p_alpha - q_alpha;
}

double Driver::E_Step(vec* N, mat* F, cube* S, const mat& buf)
{
	double Lower_bound_All = 0;
	double Lower_bound_Mpi = 0;
	if (N) *N = zeros<vec>(K);
	if (F) *F = zeros<mat>(D, K);
	if (S) *S = zeros<cube>(D, D, K);

	vec N_Mpi = zeros<vec>(K);
	mat Ex_Mpi = zeros<mat>(D, K);
	cube S_Mpi = zeros<cube>(D, D, K);

	vec N_All = zeros<vec>(K);
	mat Ex_All = zeros<mat>(D, K);
	cube S_All = zeros<cube>(D, D, K);

	// openMP for Multi threading

#pragma omp parallel for
	for (int i = 0; i < buf.n_cols; i += 1)
	{
		vec rnk;
		vec feat = buf.col(i);
		mat feat_featT = feat * feat.t();
		Compute_rnk(rnk, feat);
		double lower_bound_part = 0;
		mat tmpF = zeros<mat>(D, K);
		cube tmpS = zeros<cube>(D, D, K);
		for (int k = 0; k < K; k++)
		{
			if (rnk(k) > dstatsmin)
			{
				lower_bound_part += rnk(k) * log(rnk(k));
				tmpF.col(k) = rnk(k) * feat;
				tmpS.slice(k) = rnk(k) * feat_featT;
			}
			else
			{
				rnk(k) = 0.0;
			}
		}
#pragma omp critical
		{
			Lower_bound_All += lower_bound_part;
			N_All += rnk;
			Ex_All += tmpF;
			S_All += tmpS;
		}
	}
	Lower_bound_Mpi = Lower_bound_All;
	N_Mpi = N_All;
	Ex_Mpi = Ex_All;
	S_Mpi = S_All;

	for (int k = 0; k < K; k++)
	{
		(*N)(k) = N_Mpi(k);
		vec Exk = Ex_Mpi.col(k) / N_Mpi(k);
		(*F).col(k) = Exk;
		(*S).slice(k) = S_Mpi.slice(k) / N_Mpi(k) - Exk * Exk.t();
	}
	vWeight = (*N) / buf.n_cols;
	return Lower_bound_Mpi;
}

bool Driver::M_Step(vec& N, mat& F, cube& S, mat& buf)
{
	for (int k = 0; k < K; k++)
	{
		if (vWeight(k) < dWeightmin)
		{
			continue;
		}
		double nk = N(k);
		vec Exk = F.col(k);
		mat Sk = S.slice(k);

		vbeta(k) = vbeta0 + nk;
		mM.col(k) = (vbeta0 * mM0 + nk * Exk) / vbeta(k);
		vec vdmM = Exk - mM0;
		mInvVar.slice(k) = minvW0 + nk * Sk + vbeta0 * nk / vbeta(k) * (vdmM * vdmM.t());
		lndetW(k) = log_det_inv(mInvVar.slice(k));
		vNu(k) = vNu0 + nk;
		vgamma1(k) = 1 + nk;
		double vgammatmp = 0.0;
		for (int i = k + 1; i < K; i++)
		{
			vgammatmp += N(i);
		}
		vgamma2(k) = E_alpha + vgammatmp;

		alpha_omega1 = alpha_s1 + K;
		alpha_omega2 = alpha_s2;
		for (int i = 0; i < K; i++)
		{
			alpha_omega2 -= vE_1_V(i);
		}
	}
	return true;
}

bool Driver::Train()
{
	
	int i = 0, j = 0, k = 0, m = 0, n = 0;
	Dm_Last = -1e100;
	int iEM_Now = 1;
	field<mat> feature_set;
	ifstream  fp(sFeatureList.c_str());
	if (!fp)
	{
		return false;
	}
	string buf;
	vector <string> str_list;
	while (getline(fp, buf))
	{
		str_list.push_back(buf);
	}
	feature_set.set_size(str_list.size());
#pragma omp parallel for
	for (int i = 0; i < feature_set.n_elem; i++)
	{
		feature_set[i].load(str_list[i]);
	}
	if (feature_set.n_elem == 0)
	{
		return false;
	}
	mat labelTrainMat = feature_set[0];
	for (int i = 1; i < feature_set.n_elem; i++)
	{
		labelTrainMat = join_cols(labelTrainMat, feature_set(i));
	}
	labelTrainMat = labelTrainMat.t();
	
	InitVariables(labelTrainMat);
	while (true)
	{
		Lower_bound_All = 0;
		double Lower_bound_Mpi = 0;
		vec N;
		mat F;
		cube S;
		ComputeConst();
		double q_z_at_v = E_Step(&N, &F, &S, labelTrainMat);
		Dm = ComputeCost(N, F, S, q_z_at_v);
		printf("The Lower_bound of Em IT %d at all training dataset is %lf\n", iEM_Now, Dm);
		M_Step(N, F, S, labelTrainMat);
		n_Shed = Shed_Mixture();
		if (n_Shed > 0)
		{
			printf("%d mixture deleted, The next IT's Lower_bound will not be precise enough!\n", n_Shed);
		}
		if (n_Shed >= 0)
		{
			printf("Weight of each component is");
			vWeight.t().print();
		}


		UpdateMPIVariables();
		if (abs(Dm - Dm_Last) < eps || iEM_Now >= iEM_Iternum_Max || n_Shed == -1)
		{
			break;
		}
		Dm_Last = Dm;
		iEM_Now++;
	}
	labelTrainMat.reset();
	return true;
}

bool Driver::Compute_Score()
{
	ifstream  fp(sScoreList.c_str());
	if (!fp)
	{
		return false;
	}
	string buf;
	vector<string> outputs;
	vector<string> inputs;
	stringstream sstr(weight_str);
	rowvec vweight_tmp;
	vweight_tmp.load(sstr);
	vec vweight_score = vweight_tmp.t();
	while (getline(fp, buf))
	{
		filesystem::path pathtmp(buf);
		string str = filesystem::basename(pathtmp);
		outputs.push_back(sScore_Dir + "/" + str + ".score");
		inputs.push_back(sProjectionDir + "/" + str + ".vec");
	}
	Load();
	ComputeConst();
	mat scoreMat;
	if (inputs.size() > 0)
	{
		mat tmp;
		tmp.load(inputs[0]);
		scoreMat = zeros<mat>(tmp.n_cols, inputs.size());
	}

#pragma omp parallel for
	for (int i = 0; i < inputs.size(); i++)
	{
		vec scoreVec;
		mat ProjectMat;
		ProjectMat.load(inputs[i]);
		Score(ProjectMat, vweight_score, scoreVec);
#pragma omp critical
		{
			scoreMat.col(i) = scoreVec;
		}

	}
	/*vec score_array = zeros<vec>(scoreMat.n_elem);
	memcpy(score_array.memptr(), scoreMat.memptr(), scoreMat.n_elem *sizeof(double));
	double score_mean = mean(score_array);
	double score_var = var(score_array);
	scoreMat = (scoreMat - score_mean) / score_var;*/
	for (int i = 0; i < inputs.size(); i++)
	{
		SaveScore(outputs[i], labelset, scoreMat.col(i), true);
	}
	return true;
}

double Driver::Score_Euclidean(vec& Hvector, vec& Hprob)
{
	double score = 0;
	mat kvec = reshape(Hvector, D, K);
	if (Hprob.n_elem != D)
	{
		for (int k = 0; k < K; k++)
		{
			score += norm((kvec.col(k) - mM.col(k)), 2);
		}
	}
	else
	{
		for (int k = 0; k < K; k++)
		{
			score += norm(Hprob % (kvec.col(k) - mM.col(k)), 2);
		}
	}
	return score;
}

double Driver::Score_Cosine(vec& Hvector, vec& Hprob)
{
	double score = 0;
	if (D == 1)
	{
		return Score_Euclidean(Hvector, Hprob);
	}
	else
	{
		mat kvec = reshape(Hvector, D, K);
		for (int k = 0; k < K; k++)
		{
			score -= Hprob(k) * dot(kvec.col(k), mM.col(k)) / (norm(kvec.col(k), 2) * norm(mM.col(k), 2));
		}
	}
	return score;
}

double Driver::E_ln_L(double ln_det_W, double v, int D) const
{
	double sum_part1 = 0.0;
	for (int i = 1; i <= D; i++)
	{
		sum_part1 += gsl_sf_psi(0.5 * (v + 1 - i));
	}
	return (sum_part1 + D * log(2) + ln_det_W);
}

double Driver::H(double ln_det_W, double v, int D) const
{
	double dvalue = (Minus_ln_B(ln_det_W, v, D) - 0.5 * (v - D - 1) * E_ln_L(ln_det_W, v, D) + 0.5 * (v + D));
	return dvalue;
}

double Driver::Minus_ln_B(double ln_det_W, double v, int D) const
{
	double sum_part1 = 0.0;
	for (int i = 1; i <= D; i++)
	{
		sum_part1 += gsl_sf_lngamma(0.5 * (v + 1 - i));
	}
	double dvalue = (0.5 * v * ln_det_W + 0.5 * v * D * log(2) + 0.25 * D * (D - 1) * log(M_PI) + sum_part1);
	return dvalue;
}

bool Driver::Score(mat& HMatrix, vec & weight_vec, vec& scoreVec)
{
	vec vcos = zeros<vec>(HMatrix.n_cols);
	vec vdis = zeros<vec>(HMatrix.n_cols);

	for (int i = 0; i < vdis.n_elem; i++)
	{
		vec tmpvec = HMatrix.col(i);
		vdis(i) = Score_Euclidean(tmpvec, weight_vec);
	}
	//Norm_01(vdis);
	scoreVec = vdis;
	return true;
}

double Driver::log_det_inv(mat& orimat) const
{
	/*mat U = chol(orimat);
	return -2 * sum(log((U.diag())));*/
	double lndetW, sign;
	log_det(lndetW, sign, orimat);
	return -lndetW;
}

double Driver::Mahalanobis_Distance(const vec& feat, const vec& mean, const mat& InvVar) const
{
	vec vtmp = feat - mean;
	return dot(vtmp, solve(InvVar, vtmp));
}

bool Driver::UpdateMPIVariables()
{

	return true;
}

bool Driver::Norm_01(vec& scoreVec)
{
	scoreVec = (scoreVec - mean(scoreVec)) / var(scoreVec);
	double dmax = scoreVec.max();
	double dmin = scoreVec.min();
	if (dmax == dmin)
	{
		scoreVec.fill(0.0);
		return false;
	}
	for (int i = 0; i < scoreVec.n_elem; i++)
	{
		scoreVec(i) = (scoreVec(i) - dmin) / (dmax - dmin);
	}
	return true;
}

int Driver::Shed_Mixture()
{
	int shed_num = 0;
	uvec delvec = arma::sort(find(vWeight < dWeightmin), 1);
	if (delvec.n_elem == 0)
	{
		return 0;
	}
	else if (delvec.n_elem > 0 && K - delvec.n_elem < cluster_min)
	{
		shed_num = K - cluster_min;
		K -= shed_num;
		for (int i = 0; i < shed_num; i++)
		{
			vWeight.shed_row(delvec(i));
			mM.shed_col(delvec(i));
			mInvVar.shed_slice(delvec(i));
			lndetW.shed_row(delvec(i));
			vgamma1.shed_row(delvec(i));
			vgamma2.shed_row(delvec(i));
			vbeta.shed_row(delvec(i));
			vNu.shed_row(delvec(i));
		}
		return -1;
	}
	else
	{
		shed_num = delvec.n_elem;
		K -= shed_num;
		for (int i = 0; i < shed_num; i++)
		{
			vWeight.shed_row(delvec(i));
			mM.shed_col(delvec(i));
			mInvVar.shed_slice(delvec(i));
			lndetW.shed_row(delvec(i));
			vgamma1.shed_row(delvec(i));
			vgamma2.shed_row(delvec(i));
			vbeta.shed_row(delvec(i));
			vNu.shed_row(delvec(i));
		}
		return shed_num;
	}

}
