#include "Driver_Tools.h"
#include "Driver.h"
#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <iterator>
#include <omp.h>
namespace po = boost::program_options;
using namespace std;
Model_Task task;

enum SYS_CODE
{
	SYS_SUCCESS = 0,
	SYS_ERROR_TRAIN,
	SYS_ERROR_Projection,
	SYS_ERROR_SCORE,
	SYS_ERROR_INVALID_INPUT,
	SYS_ERROR_CONFIG,
	SYS_ERROR_CLUSTER,
	SYS_ERROR_MODEL_OPT,
	SYS_ERROR_MEMORY,
	SYS_ERROR_FILE_OPT,
	SYS_ERROR_INIT,
	SYS_ERROR_EXIT,
};


bool LoadFeatureList(string filelist, field<mat> &feature_set)
{
	ifstream  fp(filelist.c_str());
	if (!fp)
	{
		return false;
	}
	string buf;
	vector <string> str_list;
	while (getline(fp, buf))
	{
		str_list.push_back(buf);
	}
	feature_set.set_size(str_list.size());
	for (int i = 0; i < feature_set.n_elem; i++)
	{
		feature_set[i].load(str_list[i]);
	}
	return true;
}


template<class T>
SYS_CODE System_Init(T &model, int ac, char **av)
{
	if (model.Init(ac, av))
	{

		return SYS_SUCCESS;
	}
	else
	{
		printf("Init Fail!\n");
		return SYS_ERROR_INIT;
	}
}

template<class T>
SYS_CODE System_Exit(T &model)
{
	if (model.Exit())
	{
		return SYS_SUCCESS;
	}
	else
	{
		return SYS_ERROR_EXIT;
	}
}

template<class T>
SYS_CODE System_Cluster(T &model)
{
	if (model.Compute_Cluster())
	{
		return SYS_SUCCESS;
	}
	else
	{
		return SYS_ERROR_CLUSTER;
	}
}

template<class T>
SYS_CODE System_Projection(T &model)
{
	if (model.Compute_Project())
	{
		return SYS_SUCCESS;
	}
	else
	{
		return SYS_ERROR_Projection;
	}
}

template<class T>
SYS_CODE System_Train(T &model)
{
	if (model.PreTrain()&&model.Train()&& model.Save())
	{
		return SYS_SUCCESS;
	}
	else
	{
		return SYS_ERROR_TRAIN;
	}
}

template<class T>
SYS_CODE System_Score(T &model)
{
	if (model.Compute_Score())
	{
		return SYS_SUCCESS;
	}
	else
	{
		return SYS_ERROR_SCORE;
	}
}

bool Driver_Run(int ac,char **av)
{
	Run_Mode mode;
	Driver dpm;
	mat labelTrainMat; //matrix for model training
	mat labelProjectMat; //matrix for Project
	mat labelScoreMat; //matrix for score
	vector <string> labelset;
	vec scoreVec; //vector for score
	string sModelPath;
	string sProjectionDir;
	string sFeatureList, sLabelFile, sPathWayFile, sModel, sRunModel, SWorkMode, sScore_Dir, sCluster_Result;
	bool bDoKmeans = false;
	double dcutoff;
	string config_file;
	string weight_str;
	labelset.clear();
	checkDir();
	if (SYS_SUCCESS != System_Init(dpm, ac, av))
	{
		return false;
	}
	omp_set_num_threads(dpm.threadnum);// thread number 
	switch (dpm.mode)
	{
	case R_TRAIN:
		System_Train(dpm);
		break;
	case R_PROJECTION:
		System_Projection(dpm);
		break;
	case R_SCORE:
		System_Score(dpm);
		break;
	case R_CLUSTER:
		System_Cluster(dpm);
		break;
	case R_ALL:
		System_Train(dpm);
		System_Cluster(dpm);
		System_Projection(dpm);
		System_Score(dpm);
		break;
	default:
		break;
	}
	if (SYS_SUCCESS != System_Exit(dpm))
	{
		return false;
	}
	return true;
}

